Class = require "hump.class"
Vector = require "hump.vector"
settings = require "settings"
managers = require "managers"
weapons = require "weapons"


Actor = Class{
    type = 'Actor',

    init = function(self, position, img, radius)
        self.radius = radius
        self.img = img
        self.scale = radius * 2.0 / img:getWidth()
        self.width = img:getWidth()
        self.height = img:getHeight()
        self.rotation = 0
        self.orientation = -math.pi/2
        managers.objects.add(self)

        self:initPhysics(position)
    end;

    initPhysics = function(self, position)
        self.physics = {}
        self.physics.body = love.physics.newBody(managers.world, position.x, position.y, "dynamic")
        self.physics.body:setLinearDamping(10)
        self.physics.body:setAngularDamping(10)
        self.physics.shape = love.physics.newCircleShape(self.radius)
        self.physics.fixture = love.physics.newFixture(self.physics.body, self.physics.shape, 1)
        self.physics.fixture:setRestitution(0)
        self.physics.fixture:setUserData(self)
    end;

    getPosition = function(self)
        return Vector(self:getX(), self:getY())
    end;

    getX = function(self)
        return self.physics.body:getX()
    end;

    getY = function(self)
        return self.physics.body:getY()
    end;

    setX = function(self, value)
        return self.physics.body:setX(value)
    end;

    setY = function(self, value)
        return self.physics.body:setY(value)
    end;

    remove = function(self)
        managers.objects.remove(self)
    end;

    update = function(self, dt)
    end;

    draw = function(self)
        love.graphics.push()
        
        love.graphics.translate(self:getX(), self:getY())
        if self.rotation then
            love.graphics.rotate(self.rotation)
        end
        love.graphics.scale(self.scale)
        love.graphics.translate(-self:getX(), -self:getY())
        
        love.graphics.draw(self.img, self:getX() - self.width/2, self:getY() - self.height/2)
        --self.physics.shape:getRadius()
        love.graphics.pop()
    end;
}


Destroyable = Class{__includes = Actor;
    type = 'Destroyable',
    
    hit = function(self, value)
        if self.lastDamage then
            local now = os.time()
            if now - self.lastDamage <= 1 then
                return false
            end
            self.lastDamage = now
        end
        self.health = self.health - value
        return true
    end;

    heal = function(self, value)
        self.health = self.health + value
        if self.health > self.maxHealth then
            self.health = self.maxHealth
        end
    end;

    isDead = function(self)
        return self.health <= 0
    end;

    kill = function(self)
        self:remove()
    end;
}


Movable = Class{__includes = Actor;
    type = 'Movable',

    init = function(self, position, accel, radius, img)
        Actor.init(self, position, img, radius)
        self.accel = accel
    end;

    update = function(self, dt)
        local vel = Vector(self.physics.body:getLinearVelocity())
        if not vel:isZero() then
            self.rotation = vel:angleTo() - self.orientation
        end
    end;
}

local felled_tree = managers.tiles.newQuad(3, 20)

Floor = Class{
    type = 'Floor',
    quads = {managers.tiles.newQuad(0, 20), managers.tiles.newQuad(1, 20), felled_tree},

    init = function(self, x, y)
        self.x = x
        self.y = y
        self.quad = Floor.getQuad()
    end;

    update = function(self, dt)
    end;

    getQuad = function()
        return Floor.quads[math.random(1, 3)]
    end;

    draw = function(self)
    end;
}


Wall = Class{__includes = Destroyable;
    type = 'Wall',
    quads = {managers.tiles.newQuad(2, 20), managers.tiles.newQuad(4, 20)},
    shape = love.physics.newRectangleShape(settings.world.tileWorldSize, settings.world.tileWorldSize),

    init = function(self, x, y)
        self.x = x
        self.y = y
        self.quad = Wall.getQuad()

        self.maxHealth = 3
        self.health = self.maxHealth

        managers.objects.add(self)
        local half = settings.world.tileWorldSize / 2
        self:initPhysics(x - half, y - half)
    end;

    initPhysics = function(self, x, y)
        self.physics = {}
        self.physics.body = love.physics.newBody(managers.world, x, y, "static")
        self.physics.shape = self.shape
        self.physics.fixture = love.physics.newFixture(self.physics.body, self.physics.shape, 1)
        self.physics.fixture:setRestitution(0)
        self.physics.fixture:setUserData(self)
    end;

    remove = function(self)
        managers.objects.remove(self)
    end;

    kill = function(self)
        self:remove()
        self.quad = felled_tree
    end;

    update = function(self, dt)
    end;

    draw = function(self)
    end;

    getQuad = function()
        return Wall.quads[math.random(1, 2)]
    end;
}


function getPlayerImpulse()
    local newVel = Vector()
    if love.keyboard.isDown("left") or love.keyboard.isDown("a") then
        newVel = newVel + Vector(-1, 0)
    end
    if love.keyboard.isDown("right") or love.keyboard.isDown("d")  then
        newVel = newVel + Vector(1, 0)
    end
    if love.keyboard.isDown("up") or love.keyboard.isDown("w")  then
        newVel = newVel + Vector(0, -1)
    end
    if love.keyboard.isDown("down") or love.keyboard.isDown("s")  then
        newVel = newVel + Vector(0, 1)
    end
    return newVel:normalize_inplace()
end


return {
    Player = Class{__includes = {Movable, Destroyable};
        type = 'Player',
    
        init = function(self, position)
            local conf = settings.world.player
            Movable.init(self, position, conf.accel, conf.radius, conf.skin)
            self.maxHealth = conf.health
            self.health = self.maxHealth
            self.lastDamage = os.time()
            self.weapon = weapons.HK_MP5()
        end;

        fire = function(self, target)
            if self.weapon:doShot() then
                local impulseDir = (target - self:getPosition()):normalize_inplace()
                local bpos = self:getPosition() + impulseDir * self.radius
                self.weapon:spawnShot(actors.Bullet, bpos, impulseDir)
            end
        end;

        update = function(self, dt)
            Movable.update(self, dt)
            --[[if self.vel then
                managers.sound_manager.runningSong.unpause()
            else
                managers.sound_manager.runningSong.pause()
            end]]

            local force = self.accel * getPlayerImpulse()
            self.physics.body:applyForce(force:unpack())

            self.weapon:update(dt)
        end;
    },

    Minion = Class{__includes = {Movable, Destroyable};
        type = 'Minion',
    
        init = function(self, position, level)
            local conf = settings.world.minion
            local k

            k = level * 0.5 + level * math.random()
            accel = conf.accel * (1 + k * 0.1)

            k = level * 0.5 + level * math.random()
            local radius = conf.radius * (1 + k * 0.1)
            Movable.init(self, position, accel, radius, conf.skin)
            self.maxHealth = conf.health * k
            self.health = self.maxHealth
            self.randomTarget = nil
            self.isAggression = false
            self.damage = 1
        end;

        isCameToTarget = function(self)
            return (self.randomTarget - self:getPosition()):len() < self.radius
        end;

        getImpulse = function(self, position, rect)
            local impulse_dir = position - self:getPosition()
            if not self.isAggression then
                if impulse_dir:len() > settings.world.minion.visionRadius then
                    if self.randomTarget == nil or self:isCameToTarget() then
                        local x = (rect[3] - rect[1]) * math.random() + rect[1]
                        local y = (rect[4] - rect[2]) * math.random() + rect[2]
                        self.randomTarget = Vector(x, y)
                    end
                    impulse_dir = self.randomTarget - self:getPosition()
                else
                    self.isAggression = true
                end
            end
            return impulse_dir:normalize_inplace()
        end;

        update = function(self, dt)
            Movable.update(self, dt)

            local impulseDir = self:getImpulse(
                managers.objects.player:getPosition(),
                {0, 0, settings.world.width, settings.world.height})
            local force = self.accel * impulseDir
            self.physics.body:applyForce(force:unpack())
        end;
    },


    Bullet = Class{__includes = Actor;
        type = 'Bullet',
    
        init = function(self, position, vel)
            local conf = settings.world.bullet
            Actor.init(self, position, conf.skin, conf.radius)
            self.damage = 1
            self.lifetime = conf.lifetime
            self.physics.body:applyLinearImpulse(vel.x, vel.y)
        end;

        initPhysics = function(self, position)
            self.physics = {}
            self.physics.body = love.physics.newBody(managers.world, position.x, position.y, "dynamic")
            self.physics.shape = love.physics.newCircleShape(self.radius)
            self.physics.fixture = love.physics.newFixture(self.physics.body, self.physics.shape, 1)
            self.physics.fixture:setRestitution(0.9)
            self.physics.fixture:setUserData(self)
        end;

        update = function(self, dt)
            self.lifetime = self.lifetime - dt
            if self.lifetime <= 0  then
                self:remove()
            end
        end;
    },
    Wall = Wall,
    Floor = Floor,
}